@@ -1,28 +1,9 @@
-# Copyright (c) 2020, Riverbank Computing Limited
-# All rights reserved.
-#
-# This copy of SIP is licensed for use under the terms of the SIP License
-# Agreement.  See the file LICENSE for more details.
-#
-# This copy of SIP may also used under the terms of the GNU General Public
-# License v2 or v3 as published by the Free Software Foundation which can be
-# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
-#
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
-# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-# POSSIBILITY OF SUCH DAMAGE.
+# SPDX-License-Identifier: BSD-2-Clause
+
+# Copyright (c) 2025 Phil Thompson <phil@riverbankcomputing.com>
 
 
-from collections import OrderedDict
-from distutils.sysconfig import get_python_inc, get_python_lib
+import collections
 import os
 import packaging
 import shutil
@@ -35,12 +16,18 @@ import warnings
 from .abstract_builder import AbstractBuilder
 from .abstract_project import AbstractProject
 from .bindings import Bindings
+from .buildable import BuildableFromSources
 from .configurable import Configurable, Option
 from .exceptions import UserException
-from .module import resolve_abi_version
-from .py_versions import FIRST_SUPPORTED_MINOR, LAST_SUPPORTED_MINOR
-from .pyproject import (PyProjectException, PyProjectOptionException,
-        PyProjectUndefinedOptionException)
+from .module import get_source_version_range, parse_abi_version
+from .py_versions import OLDEST_SUPPORTED_MINOR
+from .pyproject import PyProjectException, PyProjectOptionException
+
+
+# The macOS minimum version of the latest supported version of Python.  v10.13
+# is the minimum version required by Python v3.13.
+MACOS_MIN_OS_MAJOR = 10
+MACOS_MIN_OS_MINOR = 13
 
 
 class Project(AbstractProject, Configurable):
@@ -48,9 +35,6 @@ class Project(AbstractProject, Configurable):
 
     # The configurable options.
     _options = (
-        # The ABI version that the sip module should use.
-        Option('abi_version'),
-
         # The callable that will return an Bindings instance.  This is used for
         # bindings implicitly defined in the .toml file.
         Option('bindings_factory'),
@@ -67,10 +51,6 @@ class Project(AbstractProject, Configurable):
         # The list of GUI script entry points.
         Option('gui_scripts', option_type=list),
 
-        # The minimum GLIBC version required by the project.  This is used to
-        # determine the correct platform tag to use for Linux wheels.
-        Option('minimum_glibc_version'),
-
         # The minimum macOS version required by the project.  This is used to
         # determine the correct platform tag to use for macOS wheels.
         Option('minimum_macos_version'),
@@ -79,7 +59,7 @@ class Project(AbstractProject, Configurable):
         Option('py_debug', option_type=bool),
 
         # The name of the directory containing Python.h.
-        Option('py_include_dir', default=get_python_inc()),
+        Option('py_include_dir', default=sysconfig.get_path('include')),
 
         # The name of the target Python platform.
         Option('py_platform'),
@@ -106,13 +86,14 @@ class Project(AbstractProject, Configurable):
         # The fully qualified name of the sip module.
         Option('sip_module'),
 
-        # The list of files and directories, specified as glob patterns
-        # relative to the project directory, that should be included in a
-        # wheel.  If an element of list is a string then it is a pattern and
-        # files and directories are installed in the target directory.  If an
-        # element is a 2-tuple then the first part is the pattern and the
-        # second part is the name of a sub-directory relative to the target
-        # directory where the files and directories are installed.
+        # The list of files and directories, specified as glob patterns, that
+        # should be included in a wheel.  If a pattern is relative then it is
+        # taken as being relative to the project directory.  If an element of
+        # the list is a string then it is a pattern and files and directories
+        # are installed in the target directory.  If an element is a 2-tuple
+        # then the first part is the pattern and the second part is the name of
+        # a sub-directory relative to the target directory where the files and
+        # directories are installed.
         Option('wheel_includes', option_type=list),
 
         # The user-configurable options.
@@ -122,6 +103,8 @@ class Project(AbstractProject, Configurable):
                 help="enable verbose progress messages"),
         Option('name', help="the name used in sdist and wheel file names",
                 metavar="NAME", tools=['sdist', 'wheel']),
+        Option('abi_version', help="The ABI to generate code for",
+                metavar="M[.N]"),
         Option('build_dir', help="the build directory", metavar="DIR"),
         Option('build_tag', help="the build tag to be used in the wheel name",
                 metavar="TAG", tools=['wheel']),
@@ -131,11 +114,25 @@ class Project(AbstractProject, Configurable):
                 help="disable the use of manylinux in the platform tag used "
                         "in the wheel name",
                 tools=['wheel']),
-        Option('target_dir', default=get_python_lib(plat_specific=1),
+        Option('minimum_glibc_version',
+                help="the minimum GLIBC version to be used in the platform "
+                        "tag of Linux wheels",
+                metavar="M.N", tools=['wheel']),
+        Option('scripts_dir', default=os.path.dirname(sys.executable),
+                help="the scripts installation directory", metavar="DIR",
+                tools=['build', 'install']),
+        Option('target_dir', default=sysconfig.get_path('platlib'),
                 help="the target installation directory", metavar="DIR",
                 tools=['build', 'install']),
         Option('api_dir', help="generate a QScintilla .api file in DIR",
                 metavar="DIR"),
+        Option('compile', option_type=bool, inverted=True,
+                help="disable the compilation of the generated code",
+                tools=['build']),
+        Option('version_info', option_type=bool, inverted=True,
+                help="disable any reference to the SIP version number in "
+                        "generated code",
+                tools=['build']),
     )
 
     # The configurable options for multiple bindings.
@@ -149,19 +146,40 @@ class Project(AbstractProject, Configurable):
     def __init__(self, **kwargs):
         """ Initialise the project. """
 
-        super().__init__(**kwargs)
+        super().__init__()
 
         # The current directory should contain the .toml file.
         self.root_dir = os.getcwd()
-        self.bindings = OrderedDict()
+        self.arguments = None
+        self.bindings = collections.OrderedDict()
         self.bindings_factories = []
         self.builder = None
         self.buildables = []
         self.installables = []
 
+        # The provisional target ABI is that specified by abi-version in
+        # pyproject.toml (or on the command line).  Note that, for historical
+        # reasons, we may not have the source of the older minor versions in
+        # which case we will build against the oldest we do have.
+        self.target_abi = None
+        self._build_abi = None
+
         self._metadata_overrides = None
         self._temp_build_dir = None
 
+        self.initialise_options(kwargs)
+
+    @property
+    def all_modules_use_limited_abi(self):
+        """ True if all modules in the project use the limited ABI. """
+
+        for buildable in self.buildables:
+            if isinstance(buildable, BuildableFromSources):
+                if not buildable.uses_limited_api:
+                    return False
+
+        return True
+
     def apply_nonuser_defaults(self, tool):
         """ Set default values for non-user options that haven't been set yet.
         """
@@ -173,18 +191,22 @@ class Project(AbstractProject, Configurable):
             self.bindings_factory = self.import_callable(self.bindings_factory,
                     Bindings)
 
+        if self.py_major_version is None or self.py_minor_version is None:
+            self.py_major_version = sys.hexversion >> 24
+            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
+
         if self.builder_factory is None:
-            from .distutils_builder import DistutilsBuilder
-            self.builder_factory = DistutilsBuilder
+            if (self.py_major_version, self.py_minor_version) >= (3, 10):
+                from .setuptools_builder import SetuptoolsBuilder
+                self.builder_factory = SetuptoolsBuilder
+            else:
+                from .distutils_builder import DistutilsBuilder
+                self.builder_factory = DistutilsBuilder
         elif isinstance(self.builder_factory, str):
             # Convert the name to a callable.
             self.builder_factory = self.import_callable(self.builder_factory,
                     AbstractBuilder)
 
-        if self.py_major_version is None or self.py_minor_version is None:
-            self.py_major_version = sys.hexversion >> 24
-            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
-
         if self.py_platform is None:
             self.py_platform = sys.platform
 
@@ -196,11 +218,6 @@ class Project(AbstractProject, Configurable):
     def apply_user_defaults(self, tool):
         """ Set default values for user options that haven't been set yet. """
 
-        # If we are the backend to a 3rd-party frontend (most probably pip)
-        # then let it handle the verbosity of messages.
-        if self.verbose is None and tool == '':
-            self.verbose = True
-
         # This is only used when creating sdist and wheel files.
         if self.name is None:
             self.name = self.metadata['name']
@@ -253,6 +270,20 @@ class Project(AbstractProject, Configurable):
 
         return wheel_file
 
+    @property
+    def build_abi(self):
+        """ The ABI version to use to build against.  This may be later than
+        the target version if we don't have the source of the target version.
+        It must not be called before the target ABI has been finalised.
+        """
+
+        if self._build_abi is None:
+            major_version, minor_version = self.target_abi
+            oldest_source, _ = get_source_version_range(major_version)
+            self._build_abi = (major_version, max(minor_version, oldest_source))
+
+        return self._build_abi
+
     def get_bindings_dir(self):
         """ Return the name of the 'bindings' directory relative to the
         eventual target directory.
@@ -264,9 +295,7 @@ class Project(AbstractProject, Configurable):
         """ Return the name of the .dist-info directory for a target directory.
         """
 
-        return os.path.join(target_dir,
-                '{}-{}.dist-info'.format(self.name.replace('-', '_'),
-                self.version_str))
+        return os.path.join(target_dir, self.normalized_name + '.dist-info')
 
     def get_dunder_init(self):
         """ Return the contents of the __init__.py to install. """
@@ -303,7 +332,7 @@ class Project(AbstractProject, Configurable):
             name_parts = self.sip_module.split('.')
             del name_parts[-1]
 
-            return os.path.join(*name_parts)
+            return os.path.join(*name_parts) if name_parts else '.'
 
         return ''
 
@@ -315,34 +344,55 @@ class Project(AbstractProject, Configurable):
 
         platform_tag = sysconfig.get_platform()
 
-        if self.py_platform == 'darwin' and self.minimum_macos_version:
+        if self.py_platform == 'darwin':
             # We expect a three part tag so leave anything else unchanged.
             parts = platform_tag.split('-')
             if len(parts) == 3:
-                parts[1] = '{}.{}'.format(self.minimum_macos_version[0],
-                        self.minimum_macos_version[1])
+                version = parts[1]
 
+                if '.' in version:
+                    min_major, min_minor = version.split('.')
+                    min_major = int(min_major)
+                    min_minor = int(min_minor)
+                else:
+                    min_major = int(version)
+                    min_minor = 0
+
+                # Use the user supplied value if it is later than the platform
+                # value.
+                if self.minimum_macos_version:
+                    user_min_major = int(self.minimum_macos_version[0])
+                    user_min_minor = int(self.minimum_macos_version[1])
+
+                    if (min_major, min_minor) < (user_min_major, user_min_minor):
+                        min_major = user_min_major
+                        min_minor = user_min_minor
+
+                # Enforce a minimum macOS version for limited ABI projects.
+                if self.all_modules_use_limited_abi and (min_major, min_minor) < (MACOS_MIN_OS_MAJOR, MACOS_MIN_OS_MINOR):
+                    min_major = MACOS_MIN_OS_MAJOR
+                    min_minor = MACOS_MIN_OS_MINOR
+
+                # Enforce a minimum macOS version for arm64 binaries.
+                if parts[2] == 'arm64' and min_major < 11:
+                    min_major = 11
+                    min_minor = 0
+
+                # Reassemble the tag.
+                parts[1] = f'{min_major}.{min_minor}'
                 platform_tag = '-'.join(parts)
 
         elif self.py_platform == 'linux' and self.manylinux:
             # We expect a two part tag so leave anything else unchanged.
             parts = platform_tag.split('-')
             if len(parts) == 2:
-                if self.minimum_glibc_version > (2, 17):
-                    # PEP 600.
-                    parts[0] = 'manylinux'
-                    parts.insert(1,
-                            '{}.{}'.format(self.minimum_glibc_version[0],
-                                    self.minimum_glibc_version[1]))
-                elif self.minimum_glibc_version > (2, 12):
-                    # PEP 599.
-                    parts[0] = 'manylinux2014'
-                elif self.minimum_glibc_version > (2, 5):
-                    # PEP 571.
-                    parts[0] = 'manylinux2010'
+                if self.minimum_glibc_version:
+                    major, minor = self.minimum_glibc_version
                 else:
-                    # PEP 513.
-                    parts[0] = 'manylinux1'
+                    major, minor = 2, 5
+
+                parts[0] = 'manylinux'
+                parts.insert(1, f'{major}.{minor}')
 
                 platform_tag = '-'.join(parts)
 
@@ -368,13 +418,13 @@ class Project(AbstractProject, Configurable):
             if rd.split()[0] == sip_project_name:
                 return []
 
-        next_abi_major = int(self.abi_version.split('.')[0]) + 1
+        abi_major, abi_minor = self.target_abi
+        next_abi_major = abi_major + 1
 
-        return ['{} (>={}, <{})'.format(sip_project_name, self.abi_version,
-                next_abi_major)]
+        return [f'{sip_project_name} (>={abi_major}.{abi_minor}, <{next_abi_major})']
 
     def get_sip_distinfo_command_line(self, sip_distinfo, inventory,
-            generator=None, wheel_tag=None):
+            generator=None, wheel_tag=None, generator_version=None):
         """ Return a sequence of command line arguments to invoke sip-distinfo.
         """
 
@@ -395,6 +445,10 @@ class Project(AbstractProject, Configurable):
             args.append('--generator')
             args.append(generator)
 
+        if generator_version is not None:
+            args.append('--generator-version')
+            args.append(generator_version)
+
         if wheel_tag is not None:
             args.append('--wheel-tag')
             args.append(wheel_tag)
@@ -429,6 +483,68 @@ class Project(AbstractProject, Configurable):
         self.builder.install()
         self._remove_build_dir()
 
+    @property
+    def minimum_glibc_version(self):
+        """ The getter for the minimum GLIBC version. """
+
+        return self._minimum_glibc_version
+
+    @minimum_glibc_version.setter
+    def minimum_glibc_version(self, value):
+        """ The setter for the minimum GLIBC version. """
+
+        # Handle the initial creation of the option value.
+        if value is None and not hasattr(self, '_minimum_glibc_version'):
+            self._minimum_glibc_version = None
+            return
+
+        # Make sure any minimum GLIBC version is valid and convert it to a
+        # 2-tuple.
+        if value:
+            try:
+                value = self._convert_major_minor(value)
+            except ValueError:
+                raise PyProjectOptionException('minimum-glibc-version',
+                        "'{0}' is an invalid GLIBC version number".format(
+                                value),
+                        section_name='tool.sip.project')
+
+        self._minimum_glibc_version = value
+
+    @property
+    def minimum_macos_version(self):
+        """ The getter for the minimum macOS version. """
+
+        return self._minimum_macos_version
+
+    @minimum_macos_version.setter
+    def minimum_macos_version(self, value):
+        """ The setter for the minimum macOS version. """
+
+        # Handle the initial creation of the option value.
+        if value is None and not hasattr(self, '_minimum_macos_version'):
+            self._minimum_macos_version = None
+            return
+
+        # Make sure any minimum macOS version is valid and convert it to a
+        # 2-tuple.
+        if value:
+            try:
+                value = self._convert_major_minor(value)
+            except ValueError:
+                raise PyProjectOptionException('minimum-macos-version',
+                        "'{0}' is an invalid macOS version number".format(
+                                value),
+                        section_name='tool.sip.project')
+
+        self._minimum_macos_version = value
+
+    @property
+    def normalized_name(self):
+        """ The normalized project name including the version number. """
+
+        return self.name.replace('-', '_').lower() + '-' + self.version_str
+
     @staticmethod
     def open_for_writing(fname):
         """ Open a file for writing while handling any errors. """
@@ -503,19 +619,15 @@ class Project(AbstractProject, Configurable):
         # Set the initial configuration from the pyproject.toml file.
         self._set_initial_configuration(pyproject, tool)
 
-        # Add any tool-specific command line options for (so far unspecified)
+        # Add any tool-specific command line arguments for (so far unspecified)
         # parts of the configuration.
-        if tool != 'pep517':
-            self._configure_from_command_line(tool, tool_description)
-        else:
-            # Until pip improves it's error reporting we give the user all the
-            # help we can.
-            self.verbose = True
+        self._configure_from_arguments(tool, tool_description)
 
-        # Now that any help has been given we can report a problematic
+        # Now that any help has been given we can report a missing
         # pyproject.toml file.
-        if pyproject.toml_error:
-            raise PyProjectException(pyproject.toml_error)
+        if pyproject.pyproject is None:
+            raise PyProjectException(
+                    "there is no such file in the current directory")
 
         # Make sure the configuration is complete.
         self.apply_user_defaults(tool)
@@ -570,6 +682,9 @@ class Project(AbstractProject, Configurable):
             if not b.is_buildable():
                 del self.bindings[b.name]
 
+        if len(self.bindings) == 0:
+            raise UserException("There are no bindings that can be built")
+
     def verify_configuration(self, tool):
         """ Verify that the configuration is complete and consistent. """
 
@@ -579,49 +694,25 @@ class Project(AbstractProject, Configurable):
                     "'{0}' must begin with a digit".format(self.build_tag),
                     section_name='tool.sip.project')
 
-        # Make sure any minimum GLIBC version is valid and convert it to a
-        # 2-tuple.
-        if self.minimum_glibc_version:
-            try:
-                self.minimum_glibc_version = self._convert_major_minor(
-                        self.minimum_glibc_version)
-            except ValueError:
-                raise PyProjectOptionException('minimum-glibc-version',
-                        "'{0}' is an invalid GLIBC version number".format(
-                                self.minimum_glibc_version),
-                        section_name='tool.sip.project')
-        else:
-            self.minimum_glibc_version = (2, 5)
-
-        # Make sure any minimum macOS version is valid and convert it to a
-        # 2-tuple.
-        if self.minimum_macos_version:
-            try:
-                self.minimum_macos_version = self._convert_major_minor(
-                        self.minimum_macos_version)
-            except ValueError:
-                raise PyProjectOptionException('minimum-macos-version',
-                        "'{0}' is an invalid macOS version number".format(
-                                self.minimum_macos_version),
-                        section_name='tool.sip.project')
-
         # Make sure relevent paths are absolute and use native separators.
         self.sip_files_dir = self.project_path(self.sip_files_dir)
         self.sip_include_dirs = [self.project_path(d)
                 for d in self.sip_include_dirs]
 
-        # Make sure we support the targeted version of Python.
+        # Check that the targeted version of Python isn't too old.  We hope
+        # that we will support newer versions automatically, but it's not
+        # guaranteed.
         py_version = (self.py_major_version, self.py_minor_version)
-        first_version = (3, FIRST_SUPPORTED_MINOR)
-        last_version = (3, LAST_SUPPORTED_MINOR)
+        first_version = (3, OLDEST_SUPPORTED_MINOR)
 
-        if py_version < first_version or py_version > last_version:
+        if py_version < first_version or self.py_major_version > 3:
             raise UserException(
                     "Python v{}.{} is not supported".format(
                             self.py_major_version, self.py_minor_version))
 
-        # Make sure we have a valid ABI version.
-        self.abi_version = resolve_abi_version(self.abi_version)
+        # Get the provisional ABI version to target.
+        if self.abi_version:
+            self.target_abi = parse_abi_version(self.abi_version)
 
         # Checks for standalone projects.
         if tool in Option.BUILD_TOOLS and not self.sip_module:
@@ -667,13 +758,14 @@ class Project(AbstractProject, Configurable):
         for bindings in self.bindings.values():
             bindings.verify_configuration(tool)
 
-    def _configure_from_command_line(self, tool, tool_description):
-        """ Update the configuration from the user supplied command line. """
+    def _configure_from_arguments(self, tool, tool_description):
+        """ Update the configuration from any user supplied arguments. """
 
         from argparse import SUPPRESS
         from .argument_parser import ArgumentParser
 
-        parser = ArgumentParser(tool_description, argument_default=SUPPRESS)
+        parser = ArgumentParser(tool_description, build_tool=True,
+                argument_default=SUPPRESS)
 
         # Add the user configurable options to the parser.
         all_options = {}
@@ -694,7 +786,7 @@ class Project(AbstractProject, Configurable):
             bindings.add_command_line_options(parser, tool, all_options)
 
         # Parse the arguments and update the corresponding configurables.
-        args = parser.parse_args()
+        args = parser.parse_args(self.arguments)
 
         for option, configurables in all_options.items():
             for configurable in configurables:
@@ -756,28 +848,22 @@ class Project(AbstractProject, Configurable):
         self.metadata = pyproject.get_metadata()
         self._metadata_overrides = self.get_metadata_overrides()
         self.metadata.update(self._metadata_overrides)
+
+        # Get the version string and convert it to a number.
         self.version_str = self.metadata['version']
 
-        # Convert the version as a string to number.
-        base_version = packaging.version.parse(self.version_str).base_version
-        base_version = base_version.split('.')
+        try:
+            version = packaging.version.parse(self.version_str)
+        except:
+            raise PyProjectOptionException('version',
+                    "'{0}' is an invalid version number".format(
+                            self.version_str),
+                    section_name='tool.sip.metadata')
 
-        while len(base_version) < 3:
-            base_version.append('0')
+        self.version = version.major << 16 | version.minor << 8 | version.micro
 
-        version = 0
-        for part in base_version:
-            version <<= 8
-
-            try:
-                version += int(part)
-            except ValueError:
-                raise PyProjectOptionException('version',
-                        "'{0}' is an invalid version number".format(
-                                self.version_str),
-                        section_name='tool.sip.metadata')
-
-        self.version = version
+        # Get the limited ABI version string.
+        self.limited_abi_version_str = self._get_limited_abi_version_str()
 
         # Configure the project.
         self.configure(pyproject, 'tool.sip.project', tool)
@@ -804,3 +890,28 @@ class Project(AbstractProject, Configurable):
         for bindings in self.bindings.values():
             bindings.configure(pyproject, 'tool.sip.bindings.' + bindings.name,
                     tool)
+
+    def _get_limited_abi_version_str(self):
+        """ Get the version of the limited ABI to be used. """
+
+        try:
+            # The version of the ABI to use is taken from the project metadata.
+            spec_set = packaging.specifiers.SpecifierSet(
+                    self.metadata['requires-python'])
+
+            # Find the oldest Python version that satisfies the requirement.
+            # The 100 is an arbitrary upper bound.
+            min_req_version = next(
+                    spec_set.filter((f'3.{v}' for v in range(100))))
+
+            min_req_version = packaging.version.parse(min_req_version)
+            limited_abi_version_str = '0x{0:02x}{1:02x}{2:02x}00'.format(
+                    min_req_version.major,
+                    min_req_version.minor,
+                    min_req_version.micro)
+        except Exception as e:
+            # Default to the oldest version of Python we support.
+            limited_abi_version_str = '0x03{0:02x}0000'.format(
+                    OLDEST_SUPPORTED_MINOR)
+
+        return limited_abi_version_str
